import cv2
import numpy as np
import math

class HumanDetector():

    def __init__(self):

        import mediapipe as mp

        self.mp_face_detection = mp.solutions.face_detection
        self.mp_drawing = mp.solutions.drawing_utils
        self.mp_pose = mp.solutions.pose
        self.mp_drawing_styles = mp.solutions.drawing_styles

        
        self.mp_hands = mp.solutions.hands
        self.mp_drawing = mp.solutions.drawing_utils
        # Initialize MediaPipe Hands.
        self.hands = self.mp_hands.Hands(static_image_mode=True,max_num_hands=2,min_detection_confidence=0.4)
        self.mp_pose = mp.solutions.pose
        self.pose = self.mp_pose.Pose(min_detection_confidence=0.5, min_tracking_confidence=0.5)
        self.face_detection = self.mp_face_detection.FaceDetection(model_selection=1, min_detection_confidence=0.5)
        self.skeleton_detection = self.mp_pose.Pose(static_image_mode=True,model_complexity=2,enable_segmentation=True,min_detection_confidence=0.5)


        self.list_points = ["WRIST",
          "THUMB_CMC",
          "THUMB_MCP" ,
          "THUMB_IP",
          "THUMB_TIP",
          "INDEX_FINGER_MCP",
          "INDEX_FINGER_PIP",
          "INDEX_FINGER_DIP",
          "INDEX_FINGER_TIP",
          "MIDDLE_FINGER_MCP",
          "MIDDLE_FINGER_PIP",
          "MIDDLE_FINGER_DIP",
          "MIDDLE_FINGER_TIP",
          "RING_FINGER_MCP",
          "RING_FINGER_PIP",
          "RING_FINGER_DIP",
          "RING_FINGER_TIP",
          "PINKY_MCP",
          "PINKY_PIP",
          "PINKY_DIP",
          "PINKY_TIP"]

        self.dict_points = {"WRIST":{},"THUMB_CMC":{},"THUMB_MCP":{},
          "THUMB_IP":{},"THUMB_TIP":{},
          "INDEX_FINGER_MCP":{},"INDEX_FINGER_PIP":{},
          "INDEX_FINGER_DIP":{},"INDEX_FINGER_TIP":{},
          "MIDDLE_FINGER_MCP":{},"MIDDLE_FINGER_PIP":{},
          "MIDDLE_FINGER_DIP":{},"MIDDLE_FINGER_TIP":{},
          "RING_FINGER_MCP":{},"RING_FINGER_PIP":{},
          "RING_FINGER_DIP":{},"RING_FINGER_TIP":{},
          "PINKY_MCP":{},"PINKY_PIP":{},
          "PINKY_DIP":{},"PINKY_TIP":{}}

    def setHandsParameters(self,static_image_mode=True,max_num_hands=2,min_detection_confidence=0.4):
        self.hands = self.mp_hands.Hands(static_image_mode=static_image_mode,max_num_hands=max_num_hands,min_detection_confidence=min_detection_confidence)
    def setBodyParameters(self,min_detection_confidence=0.5, min_tracking_confidence=0.5):
        self.pose = self.mp_pose.Pose(min_detection_confidence=min_detection_confidence, min_tracking_confidence=min_tracking_confidence)
    def setFaceParameters(self,model_selection=1, min_detection_confidence=0.5):
        self.face_detection = self.mp_face_detection.FaceDetection(model_selection=model_selection, min_detection_confidence=min_detection_confidence)

    
    def getBody(self,image_):
        image = image_.copy()
        image.flags.writeable = False
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        results = self.skeleton_detection.process(image)
        y_resolution, x_resolution, c = image.shape

        # Draw the pose annotation on the image.
        image.flags.writeable = True
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        self.mp_drawing.draw_landmarks(
            image,
            results.pose_landmarks,
            self.mp_pose.POSE_CONNECTIONS,
            landmark_drawing_spec=self.mp_drawing_styles.get_default_pose_landmarks_style())
        # Flip the image horizontally for a selfie-view display.

        list_pose = ["NOSE",
            "LEFT_EYE_INNER",
            "LEFT_EYE",
            "LEFT_EYE_OUTER",
            "RIGHT_EYE_INNER",
            "RIGHT_EYE",
            "RIGHT_EYE_OUTER",
            "LEFT_EAR",
            "RIGHT_EAR",
            "MOUTH_LEFT",
            "MOUTH_RIGHT",
            "LEFT_SHOULDER",
            "RIGHT_SHOULDER",
            "LEFT_ELBOW",
            "RIGHT_ELBOW",
            "LEFT_WRIST",
            "RIGHT_WRIST",
            "LEFT_PINKY",
            "RIGHT_PINKY",
            "LEFT_INDEX",
            "RIGHT_INDEX",
            "LEFT_THUMB",
            "RIGHT_THUMB",
            "LEFT_HIP",
            "RIGHT_HIP",
            "LEFT_KNEE",
            "RIGHT_KNEE",
            "LEFT_ANKLE",
            "RIGHT_ANKLE",
            "LEFT_HEEL",
            "RIGHT_HEEL",
            "LEFT_FOOT_INDEX",
            "RIGHT_FOOT_INDEX"]

        init_value = {"x":0,"y":0, "detected":False}
        pose_values = {};

        for element in list_pose:
            pose_values[element] = init_value
            #print(element)
            try:
                value = results.pose_landmarks.landmark[self.mp_pose.PoseLandmark[element]]
                value.x, value.y = self._normalized_to_pixel_coordinates(value.x, value.y, x_resolution,y_resolution)
                pose_values[element] = {"x":value.x,"y":value.y, "detected":True}
            except:
                pass
    
        return cv2.flip(image, 1), pose_values

    def getFace(self,image_):
        image = image_.copy()
        image.flags.writeable = False
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        results = self.face_detection.process(image)
        print(results)
        y_resolution, x_resolution, c = image.shape
     
        image.flags.writeable = True
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        faces = {}
        if results.detections:
            for detection in results.detections:
                data = detection.location_data
                xmin = data.relative_bounding_box.xmin
                ymin = data.relative_bounding_box.ymin
                width = data.relative_bounding_box.width
                height= data.relative_bounding_box.height
                xmin, ymin = self._normalized_to_pixel_coordinates(xmin, ymin, x_resolution,y_resolution)
                width, height = self._normalized_to_pixel_coordinates(width, height, x_resolution,y_resolution)

                faces= {detection.label_id[0]:{"xmin":xmin, "ymin":ymin, "width":width, "height":height}}

                print(detection.label_id[0])
                print(data.relative_bounding_box.xmin)
                
   

                self.mp_drawing.draw_detection(image, detection)
        
        return cv2.flip(image, 1), faces


    def _getDistance(self,elementA,elementB):
        pixelX = elementA.x - elementB.x
        pixelY = elementA.y - elementB.y

        
        distance = (pixelX**2 + pixelY**2)**.5
        
        return distance
    

    def getHandFromPoseIndexes(self,wrist, elbow, shoulder, threshold = 0.03):

        
        ratioWristElbow = 0.33
        handRectangle_x = wrist.x + ratioWristElbow * (wrist.x - elbow.x)
        handRectangle_y = wrist.y + ratioWristElbow * (wrist.y - elbow.y)
        distanceWristElbow = self._getDistance(wrist, elbow)
        distanceElbowShoulder = self._getDistance(elbow, shoulder)
        handRectangle_width = 1.5 * max(distanceWristElbow, 0.9 * distanceElbowShoulder)



        handRectangle_height = handRectangle_width;
        #x-y refers to the center --> offset to topLeft point
        handRectangle_x = handRectangle_x - handRectangle_width / 2.0
        handRectangle_y = handRectangle_y - handRectangle_height / 2.0


        return int(handRectangle_x), int(handRectangle_y), int(handRectangle_width), int(handRectangle_height)


    def _normalized_to_pixel_coordinates(self, normalized_x, normalized_y, image_width,image_height):

      x_px = min(math.floor(normalized_x * image_width), image_width - 1)
      y_px = min(math.floor(normalized_y * image_height), image_height - 1)
      
      return x_px, y_px


    def getHands(self,image_):

        image = image_.copy()
        y_resolution, x_resolution, c = image.shape
        # Convert the BGR image to RGB, flip the image around y-axis for correct 
        # handedness output and process it with MediaPipe Hands.
        results = self.hands.process(cv2.flip(cv2.cvtColor(image, cv2.COLOR_BGR2RGB), 1))
        # Print handedness (left v.s. right hand).
        # print(results.multi_handedness)

        
        self.data = {"Right": self.dict_points.copy(),"Left": self.dict_points.copy(), "l_data":False ,"r_data":False}

        annotated_image = cv2.flip(image.copy(), 1)
        k = 0
        hand_ = "left"

        
        if results.multi_hand_landmarks:

            from google.protobuf.json_format import MessageToDict
            hand_dict = []
            for idx, hand_handedness in enumerate(results.multi_handedness):
                handedness_dict = MessageToDict(hand_handedness)
                hand_dict.append(handedness_dict)

            
            for hand_landmarks in results.multi_hand_landmarks:
                j = 0
                for data_point in hand_landmarks.landmark:

                    hand_ = hand_dict[k]["classification"][0]["label"]
                    hand_value = hand_.split("\r")
                    if hand_value[0] == "Left":
                        self.data["l_data"] = True
                    else:
                        self.data["r_data"] = True
                        
                        
                    self.data[hand_value[0]][self.list_points[j]] = {'x': data_point.x*x_resolution,'y': data_point.y*y_resolution }
                    j = j + 1

                self.mp_drawing.draw_landmarks(annotated_image, hand_landmarks, self.mp_hands.HAND_CONNECTIONS)
                k = k + 1
                
        return annotated_image, self.data
 
class ObjectDetector():

    def __init__(self):
        print("ok")
